
#include "lqr.h"

#include <cassert>
#include <string>
#include <sstream>

#define XML_LQR "LQR"
#define XML_K "K"
#define XML_P "P"
#define XML_DT "DT"
#define XML_MAT "Matd"

using namespace std;

// As and Bs should be instantaneous gradients of system dynamics
// this function "integrates" them and solves the discretized 
// algebraic riccati equation.
// note that it makes some assumptions about how state is integrated.
// these assumptions will fail once more complicated systems are used
// with mixed state such as angular velocities and quaternion joints
// at that point, i will need to move the discretization out of this
// routineee.
static void 
G_LTVLQR(const vector<Matd> &As,
	 const vector<Matd> &Bs,
	 const vector<Matd> &Qs,
	 const vector<Matd> &Rs,
	 const Matd &Qend,
	 double dt,
	 vector<Matd> &Ps,
	 vector<Matd> &Ks)
{
  assert(As.size() == Bs.size());
  assert(Bs.size() == Qs.size());
  assert(Qs.size() == Rs.size());
  
  assert(false == As.empty());
  
  for (unsigned int i = 0; i < As.size(); ++i) {
    Ps.push_back(Matd(1,1));
    Ks.push_back(Matd(1,1));
  }

  int N = As.size();

  Ps[N-1] = Qend;
  
  for (int i = N - 1; i >= 0; --i) {
    const Matd &A = As[i];
    const Matd &B = Bs[i];
    const Matd &V = Ps[i];
    const Matd &Q = Qs[i];
    const Matd &R = Rs[i];
    
    Matd M = trans(B)*V*B;
    M = M + R;
    Matd temp = inv(M);

    Ks[i] = -temp*trans(B)*V*A;
    
#if 0
    cerr << "****** " << i << "****** " << endl;
    cerr << "temp " << temp << endl;
    cerr << "B " << B << endl;
    cerr << "V " << V << endl;
    cerr << "A " << A << endl;
    cerr << "K " << Ks[i] << endl;
#endif
    
    if (i > 0) {
      Ps[i-1] = Q + trans(A)*V*A - trans(A)*V*B*temp*trans(B)*V*A;
    }
  }
}

struct LQR::M_Data
{
  M_Data() : dt(0.01) {

  }

  static bool IsK(xmlNodePtr node) {
    return (XML_ELEMENT_NODE == node->type &&
	    xmlStrEqual(node->name, BAD_CAST XML_K));
  }

  static bool IsP(xmlNodePtr node) {
    return (XML_ELEMENT_NODE == node->type &&
	    xmlStrEqual(node->name, BAD_CAST XML_P));
  }

  static Matd ParseMat(xmlNodePtr root) {
    for (xmlAttrPtr props = root->properties; props; props = props->next) {
      if (xmlStrEqual(props->name, BAD_CAST XML_MAT)) {
	string matstr((const char*)XML_GET_CONTENT(props->children));

	istringstream istr(matstr);

	Matd toReturn(1,1);
	
	istr >> toReturn;
	
	return toReturn;
      }
    }
    
    assert(false);
    return Matd(1,1);
  }

  M_Data(xmlNodePtr root) : dt(10.0) {

    assert(XML_ELEMENT_NODE == root->type);
    assert(xmlStrEqual(root->name, BAD_CAST XML_LQR));

    for (xmlAttrPtr props = root->properties; props; props = props->next) {
      if (xmlStrEqual(props->name, BAD_CAST XML_DT)) {
	dt = atof((const char*)XML_GET_CONTENT(props->children));
      }
    }
    
    for (xmlNodePtr curNode = root->children; curNode; curNode=curNode->next) {
      
      if (true == IsK(curNode)) {
	Ks.push_back(ParseMat(curNode));
      }
      else if (true == IsP(curNode)) {
	Ps.push_back(ParseMat(curNode));
      }
    }
  }

  xmlNodePtr Serialize() const {
    xmlNodePtr toReturn = xmlNewNode(NULL, BAD_CAST XML_LQR);

    {
      ostringstream ostr;
      ostr << dt;
      xmlNewProp(toReturn,
		 BAD_CAST XML_DT,
		 BAD_CAST (ostr.str().c_str()));
    }

    for (unsigned int i = 0; i < Ks.size(); ++i) {
      xmlNodePtr toAdd = xmlNewNode(NULL, BAD_CAST XML_K);

      ostringstream ostr;
      ostr << Ks[i];
      xmlNewProp(toAdd,
		 BAD_CAST XML_MAT,
		 BAD_CAST (ostr.str().c_str()));

      xmlAddChild(toReturn,toAdd);
    }

    for (unsigned int i = 0; i < Ps.size(); ++i) {
      xmlNodePtr toAdd = xmlNewNode(NULL, BAD_CAST XML_P);

      ostringstream ostr;
      ostr << Ps[i];
      xmlNewProp(toAdd,
		 BAD_CAST XML_MAT,
		 BAD_CAST (ostr.str().c_str()));

      xmlAddChild(toReturn,toAdd);
    }

    return toReturn;
  }

  ~M_Data() {

  }
  
  void LTVLQR(const vector<Matd> &As,
	      const vector<Matd> &Bs,
	      const vector<Matd> &Qs,
	      const vector<Matd> &Rs,
	      const Matd &Qend,
	      double dt_)
  {
    dt = dt_;
    G_LTVLQR(As,Bs,Qs,Rs,Qend,dt,Ps,Ks);
  }

  const Matd &GetMat(double t,
		     const vector<Matd> &mats) const
  {
    assert(mats.empty() == false);

    if (t < 0) {
      t = 0;
    }

    int f = (int)floor(t/dt+0.5);

    if (f >= (int)mats.size()) {
      f = mats.size()-1;
    }
    
    return mats[f];
  }

  const Matd &GetK(double t) const
  {
    
#if 0
    cerr << "t " << t << endl;
    cerr << "dt " << dt << endl;
    cerr << "Ks[" << f << "] = " << Ks[f] << endl;
#endif

    return GetMat(t,Ks);
  }

  const Matd &GetP(double t) const
  {
    return GetMat(t,Ps);
  }

  Vecd ComputeFeedback(double t, 
		       const Vecd &state,
		       const Vecd &desiredState) const
  {

    // this is always the most confusing part of lqr.
    // in standard descriptions, there is always a 
    // negative sign in front of the K matrix.
    // when  i compute the K matrix, I include the
    // negative sign. hence, there is no negative
    // sign here.
    Vecd dx = (state-desiredState);
    Vecd toReturn = GetK(t)*dx;
    
#if 0
    cerr << "dx(" << t << ") = " << dx << endl;
    cerr << "\tfb = " << toReturn << endl;
#endif

    return toReturn;
  }

  double ComputeCostToGo(double t,
			 const Vecd &state,
			 const Vecd &desiredState) const
  {

    Vecd dx = (state-desiredState);
  
    double toReturn = 0.5*dot(dx,GetP(t)*dx);

    return toReturn;
  }

  // This solves the discrete algebraic riccati equation
  // see for example, wikipedia. currently it uses an iterative
  // method which is rather inefficient. you can solve this
  // more efficiently using Newton's method or the algorithm
  // used by dare in matlab.
  void InfiniteHorizon(const Matd &A,
		       const Matd &B,
		       const Matd &Q,
		       const Matd &R,
		       double infT,
		       double dt_) 
  {

    dt = dt_;

    Matd P = Q;
    Matd K(1,1);
    
    double t = infT;

    // the mats passed in are gradients of
    // non-linear dynamics discretized 
    // using euler integration
    
    // solve be iterating for "infinity" where
    // infinity is just a long time
    while (t >= 0) {

      Matd M = trans(B)*P*B;
      M = M + R;
      Matd temp = inv(M);
      
      K = -temp*trans(B)*P*A;

#if 0
      cerr << "P " << P << endl;
      cerr << "temp " << temp << endl;
      cerr << "K(" << t << ")" << K << endl;
#endif

      P = Q + trans(A)*(P - P*B*temp*trans(B)*P)*A;
      
      t -= dt;
    }

    Ps.push_back(P);
    Ks.push_back(K);
  }
  
  vector<Matd> Ps;
  vector<Matd> Ks;

  double dt;
};


LQR::LQR() : m_data(new M_Data())
{

}

LQR::LQR(xmlNodePtr root) : m_data(new M_Data(root))
{
  
}

LQR::~LQR()
{
  delete m_data;
}

void 
LQR::LTVLQR(const vector<Matd> &As,
	    const vector<Matd> &Bs,
	    const vector<Matd> &Qs,
	    const vector<Matd> &Rs,
	    const Matd &Qend,
	    double dt)
{
  m_data->LTVLQR(As,Bs,Qs,Rs,Qend,dt);
}

void 
LQR::InfiniteHorizon(const Matd &A,
		     const Matd &B,
		     const Matd &Q,
		     const Matd &R,
		     double infT,
		     double dt)
{
  m_data->InfiniteHorizon(A,B,Q,R,infT,dt);
}

const Matd &
LQR::GetK(double t) const
{
  return m_data->GetK(t);
}
  
Vecd 
LQR::ComputeFeedback(double t, 
		     const Vecd &state,
		     const Vecd &desiredState) const
{
  return m_data->ComputeFeedback(t,state,desiredState);
}

bool
LQR::IsRoot(xmlNodePtr node)
{
  return (XML_ELEMENT_NODE == node->type &&
	  xmlStrEqual(node->name, BAD_CAST XML_LQR));
}

xmlNodePtr
LQR::Serialize() const
{
  return m_data->Serialize();
}

double 
LQR::ComputeCostToGo(double t,
		     const Vecd &state,
		     const Vecd &desiredState) const
{
  return m_data->ComputeCostToGo(t,state,desiredState);
}

const Matd &
LQR::GetTerminalCostMatrix() const
{
	return m_data->Ps[m_data->Ps.size()-1];
}